/************************************************************************
 *   IRC - Internet Relay Chat, ircd/s_serv.c (formerly ircd/s_msg.c)
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers. 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include <string.h>
#include "h.h"

static char buf[BUFSIZE];


/*
** m_functions execute protocol messages on this server:
**
**	cptr	is always NON-NULL, pointing to a *LOCAL* client
**		structure (with an open socket connected!). This
**		identifies the physical socket where the message
**		originated (or which caused the m_function to be
**		executed--some m_functions may call others...).
**
**	sptr	is the source of the message, defined by the
**		prefix part of the message if present. If not
**		or prefix not found, then sptr==cptr.
**
**		(!IsServer(cptr)) => (cptr == sptr), because
**		prefixes are taken *only* from servers...
**
**		(IsServer(cptr))
**			(sptr == cptr) => the message didn't
**			have the prefix.
**
**			(sptr != cptr && IsServer(sptr) means
**			the prefix specified servername. (?)
**
**			(sptr != cptr && !IsServer(sptr) means
**			that message originated from a remote
**			user (not local).
**
**		combining
**
**		(!IsServer(sptr)) means that, sptr can safely
**		taken as defining the target structure of the
**		message in this server.
**	*Always* true (if 'parse' and others are working correct):
**
**	1)	sptr->from == cptr  (note: cptr->from == cptr)
**
**	2)	MyConnect(sptr) <=> sptr == cptr (e.g. sptr
**		*cannot* be a local connection, unless it's
**		actually cptr!). [MyConnect(x) should probably
**		be defined as (x == x->from) --msa ]
**
**	parc	number of variable parameter strings (if zero,
**		parv is allowed to be NULL)
**
**	parv	a NULL terminated list of parameter pointers,
**
**			parv[0], sender (prefix string), if not present
**				this points to an empty string.
**			parv[1]...parv[parc-1]
**				pointers to additional parameters
**			parv[parc] == NULL, *always*
**
**		note:	it is guaranteed that parv[0]..parv[parc-1] are all
**			non-NULL pointers.
*/
 /***********************************************************************
 * m_connect() - Added by Jto 11 Feb 1989
 ***********************************************************************/

/*
** m_connect
**	parv[0] = sender prefix
**	parv[1] = servername
**	parv[2] = port number
**	parv[3] = remote server
*/
int
m_connect (cptr, sptr, parc, parv)
     aClient *cptr, *sptr;
     int parc;
     char *parv[];
{
  int port, tmpport, retval;
  aConfItem *aconf, *cconf;
  aClient *acptr;

  if (!IsPrivileged (sptr))
    {
      sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
      return -1;
    }

  if (IsLocOp (sptr) && parc > 3)	/* Only allow LocOps to make */
    return 0;			/* local CONNECTS --SRB      */

  if (parc > 3 && MyClient (sptr))
    {
      aClient *acptr2, *acptr3;
      if (!(acptr3 = find_match_server (parv[3])))
	{
	  sendto_one (sptr, err_str (ERR_NOSUCHSERVER),
		      me.name, parv[0], parv[3]);
	  return 0;
	}

      /* Look for closest matching server */
      for (acptr2 = acptr3; acptr2 != &me; acptr2 = acptr2->serv->up)
	if (!match (parv[3], acptr2->name))
	  acptr3 = acptr2;

      parv[3] = acptr3->name;
    }

  if (hunt_server (cptr, sptr, ":%s CONNECT %s %s :%s",
		   3, parc, parv) != HUNTED_ISME)
    return 0;

  if (parc < 2 || *parv[1] == '\0')
    {
      sendto_one (sptr, err_str (ERR_NEEDMOREPARAMS),
		  me.name, parv[0], "CONNECT");
      return -1;
    }

  if ((acptr = find_server (parv[1], NULL)))
    {
      sendto_one (sptr, ":%s NOTICE %s :Connect: Server %s %s %s.",
		  me.name, parv[0], parv[1], "already exists from",
		  acptr->from->name);
      return 0;
    }

  for (aconf = conf; aconf; aconf = aconf->next)
    if (aconf->status == CONF_CONNECT_SERVER &&
	matches (parv[1], aconf->name) == 0)
      break;
  /* Checked first servernames, then try hostnames. */
  if (!aconf)
    for (aconf = conf; aconf; aconf = aconf->next)
      if (aconf->status == CONF_CONNECT_SERVER &&
	  (matches (parv[1], aconf->host) == 0 ||
	   matches (parv[1], index (aconf->host, '@') + 1) == 0))
	break;

  if (!aconf)
    {
      sendto_one (sptr,
		  "NOTICE %s :Connect: Host %s not listed in ircd.conf",
		  parv[0], parv[1]);
      return 0;
    }
  /*
     ** Get port number from user, if given. If not specified,
     ** use the default form configuration structure. If missing
     ** from there, then use the precompiled default.
   */
  tmpport = port = aconf->port;
  if (parc > 2 && !BadPtr (parv[2]))
    {
      if ((port = atoi (parv[2])) <= 0)
	{
	  sendto_one (sptr,
		      "NOTICE %s :Connect: Illegal port number", parv[0]);
	  return 0;
	}
    }
  else if (port <= 0 && (port = PORTNUM) <= 0)
    {
      sendto_one (sptr, ":%s NOTICE %s :Connect: missing port number",
		  me.name, parv[0]);
      return 0;
    }

  /*
     ** Evaluate connection rules...  If no rules found, allow the
     ** connect.  Otherwise stop with the first true rule (ie: rules
     ** are ored together.  Oper connects are effected only by D
     ** lines (CRULEALL) not d lines (CRULEAUTO).
   */
  for (cconf = conf; cconf; cconf = cconf->next)
    if ((cconf->status == CONF_CRULEALL) &&
	(matches (cconf->host, aconf->name) == 0))
      if (crule_eval (cconf->passwd))
	{
	  sendto_one (sptr,
		      "NOTICE %s :Connect: Disallowed by rule: %s",
		      parv[0], cconf->name);
	  return 0;
	}

  /*
     ** Notify all operators about remote connect requests
   */
  if (!IsAnOper (cptr))
    {
      sendto_ops_butone (NULL, &me,
			 ":%s GNOTICE :Remote CONNECT %s %s from %s",
			 me.name, parv[1], parv[2] ? parv[2] : "",
			 get_client_name (sptr, FALSE));
#if defined(USE_SYSLOG) && defined(SYSLOG_CONNECT)
      syslog (LOG_DEBUG, "CONNECT From %s : %s %d", parv[0], parv[1],
	      parv[2] ? parv[2]
#endif
	      }

	      aconf->port = port;
	      switch (retval = connect_server (aconf, sptr, NULL))
	      {
case 0:
sendto_one (sptr, ":%s NOTICE %s :*** Connecting to %s[%s].", me.name, parv[0], aconf->host, aconf->name); sendto_realops ("%s requested connect to %s[%s].", parv[0], aconf->host, aconf->name); break; case -1:
	      /* Comments already sent */
break; case -2:
sendto_one (sptr, ":%s NOTICE %s :*** Host %s is unknown.", me.name, parv[0], aconf->host); break; default:
	      sendto_one (sptr,
			  ":%s NOTICE %s :*** Connection to %s failed: %s",
			  me.name, parv[0], aconf->host, strerror (retval));}
	      aconf->port = tmpport; return 0;}
