/************************************************************************
 *   IRC - Internet Relay Chat, ircd/s_serv.c (formerly ircd/s_msg.c)
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers. 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include <string.h>
#include "h.h"

static char buf[BUFSIZE];


/*
** m_functions execute protocol messages on this server:
**
**	cptr	is always NON-NULL, pointing to a *LOCAL* client
**		structure (with an open socket connected!). This
**		identifies the physical socket where the message
**		originated (or which caused the m_function to be
**		executed--some m_functions may call others...).
**
**	sptr	is the source of the message, defined by the
**		prefix part of the message if present. If not
**		or prefix not found, then sptr==cptr.
**
**		(!IsServer(cptr)) => (cptr == sptr), because
**		prefixes are taken *only* from servers...
**
**		(IsServer(cptr))
**			(sptr == cptr) => the message didn't
**			have the prefix.
**
**			(sptr != cptr && IsServer(sptr) means
**			the prefix specified servername. (?)
**
**			(sptr != cptr && !IsServer(sptr) means
**			that message originated from a remote
**			user (not local).
**
**		combining
**
**		(!IsServer(sptr)) means that, sptr can safely
**		taken as defining the target structure of the
**		message in this server.
**	*Always* true (if 'parse' and others are working correct):
**
**	1)	sptr->from == cptr  (note: cptr->from == cptr)
**
**	2)	MyConnect(sptr) <=> sptr == cptr (e.g. sptr
**		*cannot* be a local connection, unless it's
**		actually cptr!). [MyConnect(x) should probably
**		be defined as (x == x->from) --msa ]
**
**	parc	number of variable parameter strings (if zero,
**		parv is allowed to be NULL)
**
**	parv	a NULL terminated list of parameter pointers,
**
**			parv[0], sender (prefix string), if not present
**				this points to an empty string.
**			parv[1]...parv[parc-1]
**				pointers to additional parameters
**			parv[parc] == NULL, *always*
**
**		note:	it is guaranteed that parv[0]..parv[parc-1] are all
**			non-NULL pointers.
*/
int
m_gline (aClient * cptr, aClient * sptr, int parc, char *parv[])
{
  aGline *agline, *a2gline = NULL;
  aConfItem *tmp;
  aClient *acptr;
  char *user, *host, *iphst;
  int active = -1;
#ifdef GPATH
  int logfile;
#endif /* GPATH */


  if (IsServer (cptr))
  {
    if (find_conf_host (cptr->confs, sptr->name, CONF_ULINE))
    {
      if (parc < 3 || (*parv[2] != '-' && (parc < 4 || *parv[3] == '\0')))
      {
        sendto_one (sptr, err_str (ERR_NEEDMOREPARAMS), me.name, parv[0], "GLINE");
        return 0;
      }

      if (*parv[2] == '-')	/* add mode or delete mode? */
        active = GLINE_INACTIVE;
      else
        active = GLINE_ACTIVE;

      if (*parv[2] == '+' || *parv[2] == '-')
        parv[2]++;		/* step past mode indicator */

      /* forward the message appropriately */
      if (!(acptr = find_person (parv[1], NULL)))
        sendto_serv_butone (cptr, active == GLINE_INACTIVE ? ":%s GLINE %s -%s" : ":%s GLINE %s +%s :%s",
          parv[0], parv[1], parv[2], parv[3]);	/* global! */
      else if (IsServer (acptr) || !MyConnect (acptr))
        sendto_one (acptr, active == GLINE_INACTIVE ? ":%s %s %s -%s" : ":%s %s %s +%s :%s",
          parv[0], Protocol (acptr) >= 12 ? "G" : "GLINE", parv[1], parv[2], parv[3]);	/* single destination */


      if (!(host = strchr (parv[2], '@')))
      {			/* convert user@host */
        user = "*";	/* no @'s; assume username is '*' */
        host = parv[2];
      }
      else
      {
        user = parv[2];
        *(host++) = '\0';	/* break up string at the '@' */
      }

      for (agline = gline, a2gline = NULL; agline; agline = agline->next)
      {
        if (!mycmp (agline->name, user) && !mycmp (agline->host, host))
          break;
        a2gline = agline;
      }

      if (active == GLINE_INACTIVE && agline)
      {		/* removing the gline */
        sendto_realops ("%s removing GLINE for %s@%s", parv[0], agline->name, agline->host);	/* notify opers */

#ifdef GPATH
        (void) alarm (3);	/* make a log entry */
        logfile = open (GPATH, O_WRONLY | O_APPEND | O_CREAT, S_IRUSR | S_IWUSR);
        (void) alarm (0);
        (void) sprintf (buf, "# %lu %s removing GLINE for %s@%s\n", TStime (), parv[0], agline->name, agline->host);
        (void) alarm (3);
        (void) write (logfile, buf, strlen (buf));
        (void) alarm (0);
        (void) close (logfile);
#endif /* GPATH */

        free_gline (agline, a2gline);	/* remove the gline */
      }
      else if (active != GLINE_INACTIVE)
      {		/* must be adding a gline */
        if (!agline)
        {		/* create gline */
          /* find any pre-existing K-lines */
          for (tmp = conf; tmp; tmp = tmp->next)
          {
            if ((tmp->status == CONF_KILL) && tmp->host && tmp->name)
            {
              if ((!match (tmp->name, user)) && (!match (tmp->host, host)))
                return 0;	/* found an existing K-line that matches */
            }
          }
          for (agline = gline, a2gline = NULL; agline; agline = agline->next)
          {
            if ((!match(agline->name,user)) && (!match(agline->host,host)))
              return 0;  /* found an existing G-line that matches */
            a2gline = agline; /* keep track of back pointer */
          }

          sendto_realops ("%s adding GLINE for %s@%s: %s", parv[0], user, host, parv[3]);	/* inform ops */
#ifdef GPATH
          (void) alarm (3);	/* make a log entry */
          logfile = open (GPATH, O_WRONLY | O_APPEND | O_CREAT, S_IRUSR | S_IWUSR);
          (void) alarm (0);
          (void) sprintf (buf, "# %lu %s adding GLINE for %s@%s: %s\n", TStime (), parv[0], user, host, parv[3]);
          (void) alarm (3);
          (void) write (logfile, buf, strlen (buf));
          (void) alarm (0);
          /* this can be inserted into the conf */
          (void) sprintf (buf, "K:%s:%s:%s\n", host, parv[3], user);
          (void) alarm (3);
          (void) write (logfile, buf, strlen (buf));
          (void) alarm (0);
          (void) close (logfile);
#endif /* GPATH */
          agline = make_gline (host, parv[3], user);	/* add the line */
          for (active = 0; active <= highest_fd; active++)	/* get the users! */
          {
            if ((acptr = local[active]) && !IsMe (acptr))
            {
		  find_gkill(acptr);
            }
          }
        }
      }
    }
    else
      return 0;		/* server not permitted to add a GLINE anyway */
  }
  else if (parc < 2)
  {			/* not enough args and a user...list glines */
    if (!IsOper (sptr))
    {
      sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
      return 0;
    }
    for (agline = gline, a2gline = NULL; agline; agline = agline->next)
    {
      /* display the GLINE */
      sendto_one (cptr, rpl_str (RPL_GLIST), me.name, parv[0], agline->name, agline->host, NULL, agline->reason,
                 (agline->active == GLINE_INACTIVE) ? " (Inactive)" : "");
      a2gline = agline;	/* keep the back pointer */
    }
    sendto_one (cptr, rpl_str (RPL_ENDOFGLIST), me.name, parv[0]);
  }
  return 0;
}
