/************************************************************************
 *   IRC - Internet Relay Chat, ircd/s_serv.c (formerly ircd/s_msg.c)
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers. 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include <string.h>
#include "h.h"

static	char	buf[BUFSIZE];
int max_connection_count=1,max_client_count=1;



/*
** m_functions execute protocol messages on this server:
**
**	cptr	is always NON-NULL, pointing to a *LOCAL* client
**		structure (with an open socket connected!). This
**		identifies the physical socket where the message
**		originated (or which caused the m_function to be
**		executed--some m_functions may call others...).
**
**	sptr	is the source of the message, defined by the
**		prefix part of the message if present. If not
**		or prefix not found, then sptr==cptr.
**
**		(!IsServer(cptr)) => (cptr == sptr), because
**		prefixes are taken *only* from servers...
**
**		(IsServer(cptr))
**			(sptr == cptr) => the message didn't
**			have the prefix.
**
**			(sptr != cptr && IsServer(sptr) means
**			the prefix specified servername. (?)
**
**			(sptr != cptr && !IsServer(sptr) means
**			that message originated from a remote
**			user (not local).
**
**		combining
**
**		(!IsServer(sptr)) means that, sptr can safely
**		taken as defining the target structure of the
**		message in this server.
**	*Always* true (if 'parse' and others are working correct):
**
**	1)	sptr->from == cptr  (note: cptr->from == cptr)
**
**	2)	MyConnect(sptr) <=> sptr == cptr (e.g. sptr
**		*cannot* be a local connection, unless it's
**		actually cptr!). [MyConnect(x) should probably
**		be defined as (x == x->from) --msa ]
**
**	parc	number of variable parameter strings (if zero,
**		parv is allowed to be NULL)
**
**	parv	a NULL terminated list of parameter pointers,
**
**			parv[0], sender (prefix string), if not present
**				this points to an empty string.
**			parv[1]...parv[parc-1]
**				pointers to additional parameters
**			parv[parc] == NULL, *always*
**
**		note:	it is guaranteed that parv[0]..parv[parc-1] are all
**			non-NULL pointers.
*/

/*
** m_stats
**	parv[0] = sender prefix
**	parv[1] = statistics selector (defaults to Message frequency)
**	parv[2] = server name (current server defaulted, if omitted)
**
**	Currently supported are:
**		M = Message frequency (the old stat behaviour)
**		L = Local Link statistics
**              C = Report C and N configuration lines
*/
/*
** m_stats/stats_conf
**    Report N/C-configuration lines from this server. This could
**    report other configuration lines too, but converting the
**    status back to "char" is a bit akward--not worth the code
**    it needs...
**
**    Note:   The info is reported in the order the server uses
**            it--not reversed as in ircd.conf!
**
**    Changed 17/03/2001 riss@austnet.org
**	Made All Stats except u Oper Only
**
*/

static int report_array[19][3] = {
		{ CONF_CONNECT_SERVER,    RPL_STATSCLINE, 'C'},
		{ CONF_NOCONNECT_SERVER,  RPL_STATSNLINE, 'N'},
		{ CONF_CLIENT,            RPL_STATSILINE, 'I'},
		{ CONF_KILL,              RPL_STATSKLINE, 'K'},
		{ CONF_QUARANTINED_SERVER,RPL_STATSQLINE, 'Q'},
		{ CONF_LEAF,		  RPL_STATSLLINE, 'L'},
		{ CONF_OPERATOR,	  RPL_STATSOLINE, 'O'},
		{ CONF_HUB,		  RPL_STATSHLINE, 'H'},
		{ CONF_LOCOP,		  RPL_STATSOLINE, 'o'},
		{ CONF_CRULEALL,	  RPL_STATSDLINE, 'D'},
		{ CONF_CRULEAUTO,	  RPL_STATSDLINE, 'd'},
		{ CONF_SLINES,		  RPL_STATSSLINE, 'S'},
		{ CONF_ULINE,		  RPL_STATSULINE, 'U'},
		{ CONF_TLINES,            RPL_STATSTLINE, 'T'},
		{ CONF_XLINES,		  RPL_STATSXLINE, 'X'},
		{ CONF_VLINES,		  RPL_STATSVLINE, 'V'},
		{ CONF_WLINES,		  RPL_STATSWLINE, 'W'},
		{ CONF_ZLINES,		  RPL_STATSWLINE, 'Z'},
		{ 0, 0}
				};

static	void	report_configured_links(sptr, mask)
aClient *sptr;
int	mask;
{
	static	char	null[] = "<NULL>";
	aConfItem *tmp;
	int	*p, port;
	char	c, *host, *pass, *name;
	
	for (tmp = conf; tmp; tmp = tmp->next)
		if (tmp->status & mask)
		    {
			for (p = &report_array[0][0]; *p; p += 3)
				if (*p == tmp->status)
					break;
			if (!*p)
				continue;
			c = (char)*(p+2);
			host = BadPtr(tmp->host) ? null : tmp->host;
			pass = BadPtr(tmp->passwd) ? null : tmp->passwd;
			name = BadPtr(tmp->name) ? null : tmp->name;
			port = (int)tmp->port;
			/*
			 * On K line the passwd contents can be
			/* displayed on STATS reply. 	-Vesa
			 */
			if (tmp->status == CONF_KILL)
/*				sendto_one(sptr, rpl_str(p[1]), me.name,
					   sptr->name, c, host, pass,
					   name, port, get_conf_class(tmp)); */
                        {
/* These mods are to tell the difference between the different kinds
 * of klines.  the only effect it has is in the display.  --Russell
 */
                                if (tmp->tmpconf == KLINE_PERM) c = 'K';
                                if (tmp->tmpconf == KLINE_TEMP) c = 'k';
                                sendto_one(sptr, rpl_str(p[1]), me.name,
                                           sptr->name, c,  host,
                                           index(pass, ' ') ? "*" : pass,
                                           name, port, get_conf_class(tmp));
                        }
                        else if (tmp->status == CONF_CONNECT_SERVER)
                        {
                                if (tmp->tmpconf == KLINE_PERM) c = 'C';
                                if (tmp->tmpconf == KLINE_TEMP) c = 'c';
                                sendto_one(sptr, rpl_str(p[1]), me.name,
                                           sptr->name, c,  host, name, port,
                                           get_conf_class(tmp));
                        }
                        else if (tmp->status == CONF_NOCONNECT_SERVER)
                        {
                                if (tmp->tmpconf == KLINE_PERM) c = 'N';
                                if (tmp->tmpconf == KLINE_TEMP) c = 'n';
                                sendto_one(sptr, rpl_str(p[1]), me.name,
                                           sptr->name, c,  host, name, port,
                                           get_conf_class(tmp));
                        }
                        else if (tmp->status == CONF_HUB)
                        {
                                if (tmp->tmpconf == KLINE_PERM) c = 'H';
                                if (tmp->tmpconf == KLINE_TEMP) c = 'h';
                                sendto_one(sptr, rpl_str(p[1]), me.name,
                                           sptr->name, c,  host, name, port,
                                           get_conf_class(tmp));
                        }
			/* connect rules are classless */
			else if (tmp->status & CONF_CRULE)
				sendto_one(sptr, rpl_str(p[1]), me.name,
					   sptr->name, c, host, name);
			else if (tmp->status & CONF_QUARANTINED_SERVER)
			        sendto_one(sptr, rpl_str(p[1]), me.name,
					   sptr->name, c, name, pass);
                        else if (tmp->status & CONF_TLINES)
                                sendto_one(sptr, rpl_str(p[1]), me.name,
                                           sptr->name, c, host, pass);
                        else if (tmp->status & CONF_SLINES)
                                sendto_one(sptr, rpl_str(p[1]), me.name,
                                           sptr->name, c, host, pass);
		   	else if (tmp->status & CONF_XLINES)
				sendto_one(sptr, rpl_str(p[1]), me.name,
					   sptr->name, c, host, pass);
			else if (tmp->status & CONF_VLINES)
			   	sendto_one(sptr, rpl_str(p[1]), me.name,
					   sptr->name, c, host);
		   	else if (tmp->status & CONF_WLINES)
				sendto_one(sptr, rpl_str(p[1]), me.name,
					    sptr->name, c, host);
			else if (tmp->status & CONF_ZLINES)
				sendto_one(sptr, rpl_str(p[1]), me.name,
					    sptr->name, c, host);
			else
				sendto_one(sptr, rpl_str(p[1]), me.name,
					   sptr->name, c, host, name, port,
					   get_conf_class(tmp));
		    }
	return;
}

int	m_stats(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int	parc;
char	*parv[];
    {
      	static	char	Sformat[]  = ":%s %d %s Connection SendQ SendM SendKBytes RcveM RcveKBytes :Open since";
	static	char	Lformat[]  = ":%s %d %s %s %u %u %u %u %u :%u";
	struct	Message	*mptr;
	aClient	*acptr;
	aAline	*aaline;
	aGline	*agline, *a2gline;
#ifdef P_LINES
	aPline	*apline;
#endif
	char	stat = parc > 1 ? parv[1][0] : '\0';
	Reg1	int	i;
	int	doall = 0, wilds = 0;
	char	*name;

	if (check_registered(sptr))
		return 0;

	if (hunt_server(cptr,sptr,":%s STATS %s :%s",2,parc,parv)!=HUNTED_ISME)
		return 0;

	if (parc > 2)
	    {
		name = parv[2];
		if (!mycmp(name, me.name))
			doall = 2;
		else if (matches(name, me.name) == 0)
			doall = 1;
		if (index(name, '*') || index(name, '?'))
			wilds = 1;
	    }
	else
		name = me.name;

#ifdef TELL_STATS
	if (stat != (char)0)
		sendto_realops("STATS %c by %s (%s@%s) [%s]", stat,
			sptr->name, sptr->user->username, sptr->user->host,
			sptr->user->server->name);
#endif

	switch (stat)
	{
	case 'A' : case 'a' :
		/* DL 2000-06-28 A-lines */
		if (!IsAnOper(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	break;
		}
		/* send alines */
		for (aaline = aline; aaline; aaline = aaline->next) {
		  sendto_one(sptr, rpl_str(RPL_STATSALINE), me.name,
			     sptr->name, 'A', aaline->pattern);
		}
		break;

	case 'L' : case 'l' :
		/*
		 * send info about connections which match, or all if the
		 * mask matches me.name.  Only restrictions are on those who
		 * are invisible not being visible to 'foreigners' who use
		 * a wild card based search to list it.
		 */
                if (!IsAnOper(sptr))
                {

        sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        break;
                }

	  	sendto_one(sptr, Sformat, me.name, RPL_STATSLINKINFO, parv[0]);
		for (i = 0; i <= highest_fd; i++)
		    {
			if (!(acptr = local[i]))
				continue;
			if (IsInvisible(acptr) && (doall || wilds) &&
			    !(MyConnect(sptr) && IsOper(sptr)) &&
			    !IsAnOper(acptr) && (acptr != sptr))
				continue;
			if (!doall && wilds && matches(name, acptr->name))
				continue;
			if (!(doall || wilds) && mycmp(name, acptr->name))
				continue;
			sendto_one(sptr, Lformat, me.name,
				   RPL_STATSLINKINFO, parv[0],
				   (isupper(stat)) ?
				   get_client_name(acptr, TRUE) :
				   get_client_name(acptr, FALSE),
				   (int)DBufLength(&acptr->sendQ),
				   (int)acptr->sendM, (int)acptr->sendK,
				   (int)acptr->receiveM, (int)acptr->receiveK,
				   time(NULL) - acptr->firsttime);
		    }
		break;
	case 'C' : case 'c' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

                report_configured_links(sptr, CONF_CONNECT_SERVER|
					CONF_NOCONNECT_SERVER);
		break;
	case 'G' : case 'g' :
		if (!IsAnOper(sptr))
		{
			sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
			break;
		}
		/* send glines - duplicated in m_gline */

		for (agline = gline; agline; agline = agline->next) {
		  sendto_one(sptr, rpl_str(RPL_STATSGLINE), me.name,
			     sptr->name, 'G', agline->name, agline->host,
			     NULL, agline->reason);
		}
		break;
	case 'H' : case 'h' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }
                report_configured_links(sptr, CONF_HUB|CONF_LEAF);
		break;
	case 'I' : case 'i' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		report_configured_links(sptr, CONF_CLIENT);
		break;
	case 'K' : case 'k' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		report_configured_links(sptr, CONF_KILL);
		break;
	case 'M' : case 'm' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		for (mptr = msgtab; mptr->cmd; mptr++)
			if (mptr->count)
				sendto_one(sptr, rpl_str(RPL_STATSCOMMANDS),
					   me.name, parv[0], (mptr->token)?
					   mptr->token:" ", mptr->cmd,
					   mptr->count, mptr->bytes);
		break;
	case 'o' : case 'O' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		report_configured_links(sptr, CONF_OPS);
		break;
        case 'P' : case 'p' :
#ifdef P_LINES
                 if (!IsAnOper(sptr))
                 {
#endif
                        sendto_one(sptr, err_str(ERR_NOPRIVILEGES),
                                me.name, parv[0]);
                        break;
#ifdef P_LINES
                 }
                 /* send plines */
                for (apline = pline; apline; apline = apline->next)
                {
                        if (pline->active == FALSE)
                                continue;
                        sendto_one(sptr, rpl_str(RPL_STATSPLINE), me.name,
                                sptr->name, 'P', apline->ip,
                                (int) ((apline->expire - time(NULL)) / 60));
                 }
                 break;
#endif
	case 'Q' : case 'q' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		report_configured_links(sptr, CONF_QUARANTINED_SERVER);
		break;
	case 'R' : case 'r' :
#ifdef DEBUGMODE
		send_usage(sptr,parv[0]);
#endif
		break;
	case 'D' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		report_configured_links(sptr, CONF_CRULEALL);
		break;
	case 'd' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		report_configured_links(sptr, CONF_CRULE);
		break;
	case 'S' : case 's' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		report_configured_links(sptr, CONF_SLINES);
		break;
	case 't' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		tstats(sptr, parv[0]);
		break;
	case 'T' :
                if (!IsAnOper(sptr))
                {

        sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        break;
                }

	        report_configured_links(sptr, CONF_TLINES);
		break;
	case 'U' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

                report_configured_links(sptr, CONF_ULINE);
		break;
	case 'u' :
	    {
		register time_t nowr;

		nowr = now - me.since;
		sendto_one(sptr, rpl_str(RPL_STATSUPTIME), me.name, parv[0],
			   nowr/86400, (nowr/3600)%24, (nowr/60)%60, nowr%60);
                sendto_one(sptr, rpl_str(RPL_STATSCONN), me.name, parv[0],
			   max_connection_count, max_client_count);
		break;
	    }
	case 'V': case 'v':
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		report_configured_links(sptr, CONF_VLINES);
		break;
        case 'w' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

                calc_load(sptr, parv[0]);
                break;
	case 'W' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		report_configured_links(sptr, CONF_WLINES);
		break;
	case 'X':
#ifdef	DEBUGMODE
	  	send_listinfo(sptr, parv[0]);
		break;
#endif
	case 'x':
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		report_configured_links(sptr, CONF_XLINES);
		break;
	case 'Y' : case 'y' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		report_classes(sptr);
		break;
	case 'Z' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		report_configured_links(sptr, CONF_ZLINES);
		break;
	case 'z' :
                if (!IsAnOper(sptr))
                {
        	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        	break;
                }

		count_memory(sptr, parv[0]);
		break;
	default :
		stat = '*';
		break;
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFSTATS), me.name, parv[0], stat);
	return 0;
    }


    /*
 * parv[0] = sender
 * parv[1] = host/server mask.
 * parv[2] = server to query
 */
int	 m_lusers(cptr, sptr, parc, parv)
aClient *cptr, *sptr;
int	parc;
char	*parv[];
    {
	
	int	s_count = 0, c_count = 0, u_count = 0, i_count = 0;
	int	o_count = 0, m_client = 0, m_client_local = 0, m_server = 0;
        char mydom_mask[HOSTLEN + 1];
	aClient *acptr;

	if (check_registered_user(sptr))
		return 0;

	if (parc > 2)
		if(hunt_server(cptr, sptr, ":%s LUSERS %s :%s", 2, parc, parv)
				!= HUNTED_ISME)
			return 0;

        mydom_mask[0] = '*';
        strncpy(&mydom_mask[1], DOMAINNAME, HOSTLEN - 1);

	(void)collapse(parv[1]);
	for (acptr = client; acptr; acptr = acptr->next)
	    {
		if (parc>1)
			if (!IsServer(acptr) && acptr->user)
			    {
				if (match(parv[1], acptr->user->server->name))
					continue;
			    }
			else
	      			if (match(parv[1], acptr->name))
					continue;

		switch (acptr->status)
		{
		case STAT_SERVER:
			if (MyConnect(acptr))
				m_server++;
		case STAT_ME:
			s_count++;
			break;
		case STAT_CLIENT:
			if (IsOper(acptr))
	        		o_count++;
#ifdef	SHOW_INVISIBLE_LUSERS
			if (MyConnect(acptr)) {
		  		m_client++;
				if (matches(mydom_mask, acptr->sockhost) == 0)
				  m_client_local++;
			      }
			if (!IsInvisible(acptr))
				c_count++;
			else
				i_count++;
#else
			if (MyConnect(acptr))
			    {
				if (IsInvisible(acptr))
				    {
					if (IsAnOper(sptr))
						m_client++;
				    }
				else
					m_client++;
			    }
	 		if (!IsInvisible(acptr))
				c_count++;
			else
				i_count++;
#endif
			break;
		case STAT_PING:
			break;
		default:
			u_count++;
			break;
	 	}
	     }
#ifndef	SHOW_INVISIBLE_LUSERS
	if (IsAnOper(sptr) && i_count)
#endif
	sendto_one(sptr, rpl_str(RPL_LUSERCLIENT), me.name, parv[0],
		   c_count, i_count, s_count);
#ifndef	SHOW_INVISIBLE_LUSERS
	else
		sendto_one(sptr,
			":%s %d %s :There are %d users on %d servers", me.name,
			    RPL_LUSERCLIENT, parv[0], c_count, s_count);
#endif
	if (o_count)
		sendto_one(sptr, rpl_str(RPL_LUSEROP),
			   me.name, parv[0], o_count);
	if (u_count > 0)
		sendto_one(sptr, rpl_str(RPL_LUSERUNKNOWN),
			   me.name, parv[0], u_count);
	if ((c_count = count_channels(sptr))>0)
		sendto_one(sptr, rpl_str(RPL_LUSERCHANNELS),
			   me.name, parv[0], count_channels(sptr));
	sendto_one(sptr, rpl_str(RPL_LUSERME),
		   me.name, parv[0], m_client, m_server);
        sendto_one(sptr,
	       ":%s NOTICE %s :Highest connection count: %d (%d clients)",
               me.name, parv[0], max_connection_count, max_client_count);
        if (m_client > max_client_count)
          max_client_count = m_client;
        if ((m_client + m_server) > max_connection_count) {
          max_connection_count = m_client + m_server;
          if (max_connection_count % 10 == 0)  /* only send on even tens */
            sendto_ops("Maximum connections: %d (%d clients)",
                     max_connection_count, max_client_count);
        }
        current_load_data.local_count = m_client_local;
        current_load_data.client_count = m_client;
        current_load_data.conn_count = m_client + m_server;
	return 0;
    }
