/************************************************************************
 *   IRC - Internet Relay Chat, ircd/s_serv.c (formerly ircd/s_msg.c)
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers. 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include <string.h>
#include "h.h"

static char buf[BUFSIZE];


/*
** m_functions execute protocol messages on this server:
**
**	cptr	is always NON-NULL, pointing to a *LOCAL* client
**		structure (with an open socket connected!). This
**		identifies the physical socket where the message
**		originated (or which caused the m_function to be
**		executed--some m_functions may call others...).
**
**	sptr	is the source of the message, defined by the
**		prefix part of the message if present. If not
**		or prefix not found, then sptr==cptr.
**
**		(!IsServer(cptr)) => (cptr == sptr), because
**		prefixes are taken *only* from servers...
**
**		(IsServer(cptr))
**			(sptr == cptr) => the message didn't
**			have the prefix.
**
**			(sptr != cptr && IsServer(sptr) means
**			the prefix specified servername. (?)
**
**			(sptr != cptr && !IsServer(sptr) means
**			that message originated from a remote
**			user (not local).
**
**		combining
**
**		(!IsServer(sptr)) means that, sptr can safely
**		taken as defining the target structure of the
**		message in this server.
**	*Always* true (if 'parse' and others are working correct):
**
**	1)	sptr->from == cptr  (note: cptr->from == cptr)
**
**	2)	MyConnect(sptr) <=> sptr == cptr (e.g. sptr
**		*cannot* be a local connection, unless it's
**		actually cptr!). [MyConnect(x) should probably
**		be defined as (x == x->from) --msa ]
**
**	parc	number of variable parameter strings (if zero,
**		parv is allowed to be NULL)
**
**	parv	a NULL terminated list of parameter pointers,
**
**			parv[0], sender (prefix string), if not present
**				this points to an empty string.
**			parv[1]...parv[parc-1]
**				pointers to additional parameters
**			parv[parc] == NULL, *always*
**
**		note:	it is guaranteed that parv[0]..parv[parc-1] are all
**			non-NULL pointers.
*/

 /*
    ** m_summon should be redefined to ":prefix SUMMON host user" so
    ** that "hunt_server"-function could be used for this too!!! --msa
    ** As of 2.7.1e, this was the case. -avalon
    **
    **  parv[0] = sender prefix
    **  parv[1] = user
    **  parv[2] = server
    **  parv[3] = channel (optional)
    **
  */
int
m_summon (cptr, sptr, parc, parv)
     aClient *sptr, *cptr;
     int parc;
     char *parv[];
{
  char *host, *user, *chname;
#ifdef	ENABLE_SUMMON
  char hostbuf[17], namebuf[10], linebuf[10];
#  ifdef LEAST_IDLE
  char linetmp[10], ttyname[15];	/* Ack */
  struct stat stb;
  time_t ltime = (time_t) 0;
#  endif
  int fd, flag = 0;
#endif

  if (check_registered_user (sptr))
    return 0;
  if (parc < 2 || *parv[1] == '\0')
    {
      sendto_one (sptr, err_str (ERR_NORECIPIENT),
		  me.name, parv[0], "SUMMON");
      return 0;
    }
  user = parv[1];
  host = (parc < 3 || BadPtr (parv[2])) ? me.name : parv[2];
  chname = (parc > 3) ? parv[3] : "*";
  /*
     ** Summoning someone on remote server, find out which link to
     ** use and pass the message there...
   */
  parv[1] = user;
  parv[2] = host;
  parv[3] = chname;
  parv[4] = NULL;
  if (hunt_server (cptr, sptr, ":%s SUMMON %s %s %s", 2, parc, parv) ==
      HUNTED_ISME)
    {
#ifdef ENABLE_SUMMON
      if ((fd = utmp_open ()) == -1)
	{
	  sendto_one (sptr, err_str (ERR_FILEERROR),
		      me.name, parv[0], "open", UTMP);
	  return 0;
	}
#  ifndef LEAST_IDLE
      while ((flag = utmp_read (fd, namebuf, linebuf, hostbuf,
				sizeof (hostbuf))) == 0)
	if (StrEq (namebuf, user))
	  break;
#  else
      /* use least-idle tty, not the first
       * one we find in utmp. 10/9/90 Spike@world.std.com
       * (loosely based on Jim Frost jimf@saber.com code)
       */

      while ((flag = utmp_read (fd, namebuf, linetmp, hostbuf,
				sizeof (hostbuf))) == 0)
	{
	  if (StrEq (namebuf, user))
	    {
	      (void) sprintf_irc (ttyname, "/dev/%s", linetmp);
	      if (stat (ttyname, &stb) == -1)
		{
		  sendto_one (sptr,
			      err_str (ERR_FILEERROR),
			      me.name, sptr->name, "stat", ttyname);
		  return 0;
		}
	      if (!ltime)
		{
		  ltime = stb.st_mtime;
		  (void) strcpy (linebuf, linetmp);
		}
	      else if (stb.st_mtime > ltime)	/* less idle */
		{
		  ltime = stb.st_mtime;
		  (void) strcpy (linebuf, linetmp);
		}
	    }
	}
#  endif
      (void) utmp_close (fd);
#  ifdef LEAST_IDLE
      if (ltime == 0)
#  else
      if (flag == -1)
#  endif
	sendto_one (sptr, err_str (ERR_NOLOGIN), me.name, parv[0], user);
      else
	summon (sptr, user, linebuf, chname);
#else
      sendto_one (sptr, err_str (ERR_SUMMONDISABLED), me.name, parv[0]);
#endif /* ENABLE_SUMMON */
    }
  return 0;
}
