/************************************************************************
 *   IRC - Internet Relay Chat, ircd/s_serv.c (formerly ircd/s_msg.c)
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers. 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include <string.h>
#include "h.h"

static char buf[BUFSIZE];


/*
** m_functions execute protocol messages on this server:
**
**	cptr	is always NON-NULL, pointing to a *LOCAL* client
**		structure (with an open socket connected!). This
**		identifies the physical socket where the message
**		originated (or which caused the m_function to be
**		executed--some m_functions may call others...).
**
**	sptr	is the source of the message, defined by the
**		prefix part of the message if present. If not
**		or prefix not found, then sptr==cptr.
**
**		(!IsServer(cptr)) => (cptr == sptr), because
**		prefixes are taken *only* from servers...
**
**		(IsServer(cptr))
**			(sptr == cptr) => the message didn't
**			have the prefix.
**
**			(sptr != cptr && IsServer(sptr) means
**			the prefix specified servername. (?)
**
**			(sptr != cptr && !IsServer(sptr) means
**			that message originated from a remote
**			user (not local).
**
**		combining
**
**		(!IsServer(sptr)) means that, sptr can safely
**		taken as defining the target structure of the
**		message in this server.
**	*Always* true (if 'parse' and others are working correct):
**
**	1)	sptr->from == cptr  (note: cptr->from == cptr)
**
**	2)	MyConnect(sptr) <=> sptr == cptr (e.g. sptr
**		*cannot* be a local connection, unless it's
**		actually cptr!). [MyConnect(x) should probably
**		be defined as (x == x->from) --msa ]
**
**	parc	number of variable parameter strings (if zero,
**		parv is allowed to be NULL)
**
**	parv	a NULL terminated list of parameter pointers,
**
**			parv[0], sender (prefix string), if not present
**				this points to an empty string.
**			parv[1]...parv[parc-1]
**				pointers to additional parameters
**			parv[parc] == NULL, *always*
**
**		note:	it is guaranteed that parv[0]..parv[parc-1] are all
**			non-NULL pointers.
*/
time_t TSoffset = 0;		/* Global variable; Offset of timestamps to system clock */

/*
** m_time
**	parv[0] = sender prefix
**	parv[1] = servername
*/

int seed = 131;

int
m_time (cptr, sptr, parc, parv)
     aClient *cptr, *sptr;
     int parc;
     char *parv[];
{
  if (check_registered_user (sptr))
    return 0;
  if (hunt_server (cptr, sptr, ":%s TIME :%s", 1, parc, parv) == HUNTED_ISME)
    {
      sendto_one (sptr, rpl_str (RPL_TIME), me.name, parv[0], me.name,
		  TStime (), TSoffset, date ((long) 0));
    }
  return 0;
}

/*
** m_settime
**	parv[0] = sender prefix
**	parv[1] = new time
**      parv[2] = servername
** 	parv[3] = seed
*/
m_settime (cptr, sptr, parc, parv)
     aClient *cptr, *sptr;
     int parc;
     char *parv[];
{
  time_t t;
  long int dt;
  static char tbuf[11];

  if(!IsULine(sptr))
	return 0;
	
  if (parc == 2 && MyClient (sptr))
    parv[parc++] = me.name;

  if (parc < 3)
    {
      sendto_one (sptr, err_str (ERR_NEEDMOREPARAMS),
		  me.name, parv[0], "SETTIME");
      return 0;
    }

  t = atoi (parv[1]);
  dt = TStime () - t;

  if (t < 779557906 || dt < -9000000)
    {
      sendto_one (sptr, ":%s NOTICE %s :SETTIME: Bad value", me.name,
		  parv[0]);
      return 0;
    }

  if (IsServer (sptr))
    sendto_serv_butone (cptr, ":%s SETTIME %s %s %s", parv[0], parv[1],
			parv[2], (parc < 4) ? "" : parv[3]);
  else
    {
      sprintf (tbuf, "%lu", TStime ());
      parv[1] = tbuf;
      if (hunt_server (cptr, sptr, ":%s SETTIME %s %s %s", 2, parc, parv) !=
	  HUNTED_ISME)
	return 0;
    }

  if (parc > 3)
    {
      if (!strcmp (parv[3], "rehash"))
	{
	  rehash (&me, &me, 0);
	}
      else if (atoi (parv[3]) != 0)
	{
	  seed = atoi (parv[3]);
	}
    }

  sendto_ops ("SETTIME from %s, clock is set %ld seconds %s",
	      get_client_name (sptr, FALSE), (dt < 0) ? -dt : dt,
	      (dt < 0) ? "forwards" : "backwards");
  TSoffset -= dt;
  if (IsPerson (sptr))
    sendto_one (sptr, ":%s NOTICE %s :clock is set %ld seconds %s", me.name,
		parv[0], (dt < 0) ? -dt : dt,
		(dt < 0) ? "forwards" : "backwards");
  return 0;
}

char *
militime (sec, usec)
     char *sec, *usec;
{
  struct timeval tv;
  static char timebuf[18];

  (void) gettimeofday (&tv, NULL);
  if (sec && usec)
    sprintf (timebuf, "%ld",
	     (tv.tv_sec - atoi (sec)) * 1000 + (tv.tv_usec -
						atoi (usec)) / 1000);
  else
    sprintf (timebuf, "%ld %ld", tv.tv_sec, tv.tv_usec);
  return timebuf;
}
